#include <windows.h>
#include <stdio.h>
#include <sys\stat.h>
#define MAX_SIZE 12288
void usage()
{
	printf("Usage: IncludeInBuild Project Component Value\n");
	printf("Where:\tProject is the name of the InstallShield project *\n");
	printf("\tComponent is the name of the component(s) to be included or not *\n");
	printf("\tValue is Yes or No\n");
	printf("* Enclosed in double quotes (\") if they contain spaces\n");
	printf("\nExamples:\n\tIncludeInBuild \"My Application\" \"First Component\" Yes\n");
	printf("\tIncludeInBuild \"My Application\" ISODBC* No\n");
	printf("\tIncludeInBuild \"My Application\" \"Debug Files\\*\" No\n");
	printf("\tIncludeInBuild \"My Application\" * Yes\n");

}

void HandleLargeFile(char szFileName[], char szComponent[], char szValue[], DWORD dwSizeOriginal)
{
	char *pOldStart, *pOld, *pNewStart, *pOldSave, *pNew;
	HANDLE hOldFile, hNewFile;
	HANDLE hOldMap, hNewMap;
	BOOL bYes;
	int nDelta;
	DWORD dwSizeNew;
	char szSection[256]; //how big can a component title be?
	DWORD dwError;

	if (strcmp(szValue, "Yes") == 0) {
		bYes = TRUE;
	}
	else if (strcmp(szValue, "No") == 0) {
		bYes = FALSE;
	}
	else {
		MessageBox(NULL, "The Value must be either Yes or No", "IncludeInBuild - Error", MB_OK);
		exit (-1);
	}
	
	if ((hOldFile = CreateFile(szFileName, GENERIC_READ|GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE) {
		MessageBox(NULL, "Unable to open file", "IncludeInBuild - Error", MB_OK);
		exit (-1);
	}
	if ((hOldMap = CreateFileMapping(hOldFile, NULL, PAGE_READWRITE, 0, 0, NULL)) == NULL) {
		dwError = GetLastError();
		MessageBox(NULL, "Unable to create map file", "IncludeInBuild - Error", MB_OK);
		CloseHandle(hOldFile);
		exit (-1);
	}	
	if ((pOldStart = MapViewOfFile(hOldMap, FILE_MAP_WRITE, 0, 0, 0)) == NULL) {
		MessageBox(NULL, "Unable to map file into memory", "IncludeInBuild - Error", MB_OK);
		CloseHandle(hOldMap);
		CloseHandle(hOldFile);
		exit (-1);
	}

	*(pOldStart + dwSizeOriginal) = 0; // tag end of file

	if ((pOld = strchr(szComponent, '*')) == NULL) {
		sprintf(szSection, "[%s]", szComponent);
	}
	else {
		*pOld = 0;//truncate string at '*'
		sprintf(szSection, "[%s", szComponent);
	}
	
	nDelta = 0;//let's count how many bytes we add or delete

	if ((pOld = strstr(pOldStart, szSection)) == NULL) {
		MessageBox(NULL, "Component not found", "IncludeInBuild - Error", MB_OK);
		UnmapViewOfFile(pOldStart);
		CloseHandle(hOldMap);
		CloseHandle(hOldFile);
		exit (-1);
	}

	pNewStart = (char *) GlobalAlloc(0, dwSizeOriginal + 1024);
	memcpy(pNewStart, pOldStart, (pOld - pOldStart));
	pNew = pNewStart + (pOld - pOldStart);

	for ( pOldSave = pOld; pOld != NULL; pOld = strstr(pOld, szSection) ) {
		pOld = strstr(pOld, "INCLUDEINBUILD=") + 15; // 15 = strlen(INCLUDEINBUILD=)
		memcpy(pNew, pOldSave, (pOld - pOldSave));
		pNew += (pOld - pOldSave);
		if (strncmp(pOld, szValue, strlen(szValue)) != 0) { 
			nDelta++; //increment count of changes to file 
			if (bYes) {
				//change No to Yes
				memcpy(pNew, "Yes", 3);
				pNew += 3;
				pOld += 2;
			}
			else {
				// change Yes to No
				memcpy(pNew, "No", 2);
				pNew += 2;
				pOld += 3;
			}
		}
		pOldSave = pOld;
        }
	if (nDelta != 0) { // we changed at least one
		strncpy(pNew, pOldSave, dwSizeOriginal - (pOldSave - pOldStart));
		dwSizeNew = dwSizeOriginal + ((bYes ? 1 : -1) * nDelta); // add or subtract nDelta
		if (UnmapViewOfFile(pOldStart) == 0) {
			dwError = GetLastError();
			MessageBox(NULL, "Unable to Unmap", "IncludeInBuild - Error", MB_OK);
			CloseHandle(hOldMap);
			CloseHandle(hOldFile);
			exit (-1);
		}
		if (CloseHandle(hOldMap) == 0) {
			dwError = GetLastError();
			MessageBox(NULL, "Unable to close map", "IncludeInBuild - Error", MB_OK);
			CloseHandle(hOldFile);
			exit (-1);
		}
		if (CloseHandle(hOldFile) == 0) {
			MessageBox(NULL, "Unable to close file", "IncludeInBuild - Error", MB_OK);
			exit (-1);
		}
		if ((hNewFile = CreateFile(szFileName, GENERIC_READ|GENERIC_WRITE, 0, NULL, TRUNCATE_EXISTING, FILE_ATTRIBUTE_NORMAL|FILE_FLAG_SEQUENTIAL_SCAN, NULL)) == INVALID_HANDLE_VALUE) {
			MessageBox(NULL, "Unable to re-open file", "IncludeInBuild - Error", MB_OK);
			exit (-1);
		}
		if ((hNewMap = CreateFileMapping(hNewFile, NULL, PAGE_READWRITE, 0, dwSizeNew, NULL)) == NULL) {
			dwError = GetLastError();
			MessageBox(NULL, "Unable to recreate map file", "IncludeInBuild - Error", MB_OK);
			CloseHandle(hNewFile);
			exit (-1);
		}	
		if ((pOldStart = MapViewOfFile(hNewMap, FILE_MAP_WRITE, 0, 0, 0)) == NULL) {
			MessageBox(NULL, "Unable to remap file into memory", "IncludeInBuild - Error", MB_OK);
			CloseHandle(hNewMap);
			CloseHandle(hNewFile);
			exit (-1);
		}
		memcpy(pOldStart, pNewStart, dwSizeNew);
		GlobalFree(pNewStart);
		UnmapViewOfFile(pOldStart);
		CloseHandle(hNewMap);
		CloseHandle(hNewFile);
	}	
	else {
		GlobalFree(pNewStart);
		UnmapViewOfFile(pOldStart);
		CloseHandle(hOldMap);
		CloseHandle(hOldFile);
	}
	
}

void main(int argv, char *argc[])
{
	char szFileName[_MAX_PATH];
	char szKey[_MAX_PATH];
	char szLocation[_MAX_PATH];
	char *ptr, *save, buffer[11];
	//char temp[_MAX_PATH];
	HKEY hKey;
	DWORD dwSize;
	int result;
	struct _stat statBuf;
	BOOL bIS6 = FALSE;

	if (argv < 5) { // usually called without passing in location
		if (argv < 4) {
			usage();
			exit(1);
		}
		sprintf(szKey, "SOFTWARE\\InstallShield\\InstallShield Professional");
		if (RegOpenKeyEx(HKEY_LOCAL_MACHINE, szKey, 0, KEY_ALL_ACCESS, &hKey) != ERROR_SUCCESS) {
			MessageBox(NULL, "Unable to open InstallShield key in registry", "IncludeInBuild - Error", MB_OK);
			exit (-1);
		}
		if (RegOpenKeyEx(hKey, "6.0", 0, KEY_ALL_ACCESS, &hKey) != ERROR_SUCCESS) {
			if (RegOpenKeyEx(hKey, "5.5", 0, KEY_ALL_ACCESS, &hKey) != ERROR_SUCCESS) {
				if (RegOpenKeyEx(hKey, "5.1", 0, KEY_ALL_ACCESS, &hKey) != ERROR_SUCCESS) {
					if (RegOpenKeyEx(hKey, "5.0", 0, KEY_ALL_ACCESS, &hKey) != ERROR_SUCCESS) {
						MessageBox(NULL, "Unable to determine InstallShield version from registry", "IncludeInBuild - Error", MB_OK);
						exit (-1);
					}
				}
			}
		}
		else
		{
			//v6
			bIS6 = TRUE;
		}

		if (RegOpenKeyEx(hKey, "Main", 0, KEY_ALL_ACCESS, &hKey) != ERROR_SUCCESS) {
		MessageBox(NULL, "Unable to open registry key", "IncludeInBuild - Error", MB_OK);
		exit (-1);
		}
		dwSize = sizeof(szLocation);
		
		if (bIS6)
		{
			char *pszFileName;

			if (RegQueryValueEx(hKey, "DataLink", NULL, NULL, szLocation, &dwSize) != ERROR_SUCCESS) {
				MessageBox(NULL, "Unable to read registry value", "IncludeInBuild - Error", MB_OK);
				exit (-1);
			}

			//remove filename to get directory
			pszFileName = strrchr(szLocation, '\\');
			if (pszFileName)
			{
				*pszFileName = '\0';
			}
		}
		else
		{
			//v5
			if (RegQueryValueEx(hKey, "DataFile", NULL, NULL, szLocation, &dwSize) != ERROR_SUCCESS) {
				MessageBox(NULL, "Unable to read registry value", "IncludeInBuild - Error", MB_OK);
				exit (-1);
			}
		}
	}
	else { //undocumented feature.  Can pass in location of projects.
		strcpy(szLocation, argc[4]);
	}
	sprintf(szFileName, "%s\\%s\\Component Definitions\\default.cdf", szLocation, argc[1]);
	//sprintf(temp,"Calling with (%s, INCLUDEINBUILD, %s, %s)", argc[2], argc[3], szFileName);
	//MessageBox(NULL, temp, "WritePrivateProfileString", MB_OK);
	result = _stat(szFileName, &statBuf);
	if (result != 0) {
		sprintf(szKey, "Unable to find and/or determine the size of file:\n\"%s\"", szFileName);
		MessageBox(NULL, szKey, "IncludeInBuild - Error", MB_OK);
		exit (-1);
	}
	if (statBuf.st_size > 64000) { //not sure of exact size that causes problem
		HandleLargeFile(szFileName, argc[2], argc[3], statBuf.st_size);
	}
	else { //handle most files
		if ((ptr = strchr(argc[2], '*')) == NULL) {
			WritePrivateProfileString(argc[2], "INCLUDEINBUILD", argc[3], szFileName);
		}
		else {
			*ptr = 0;//truncate string at '*'
			save = ptr = (char *)GlobalAlloc(0, MAX_SIZE);
			GetPrivateProfileString(NULL, NULL, argc[3], ptr, MAX_SIZE, szFileName);
			for ( ; *ptr != 0; ptr += strlen(ptr) + 1) {
				if (strncmp(ptr, argc[2], strlen(argc[2])) == 0) {
					GetPrivateProfileString(ptr, "INCLUDEINBUILD", argc[3], buffer, 10, szFileName);
					if (strcmp(buffer, argc[3]) != 0) {
						WritePrivateProfileString(ptr, "INCLUDEINBUILD", argc[3], szFileName);
					}
				}
			}
			GlobalFree(save);
		}
	}
}
